/******************************************************************************
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        iqs324-example-code.ino                                       *
 * @brief       IQS324 Hall Rotation Example code                             *
 * @details     This example demonstrates how to write the desired settings   *
 *              to the IQS324 to use the Hall-Rotation EV-Kit, and how        *
 *              to read the encoder interval output and touch events.         *
 *                                                                            *
 *              Data is displayed over serial communication with the          *
 *              following outputs:                                            *
 *                  - Hall Scroll with I2C interval output                    *
 *                  - Touch events                                            *
 *                  - Power Mode Switching (if enabled)                       *
 *                  - Force Communication when 'f' is sent over Serial        *
 *              When HID is enabled (DEMO_HID_ON 1), mouse scroll outputs     *
 *              will be streamed over HID (Keep Serial Monitor open)          *
 *                                                                            *
 * @author      Azoteq PTY Ltd                                                *
 * @version     v1.0.0                                                        *
 * @date        2025-03-27                                                    *
 ******************************************************************************/
#include <Arduino.h>
#include "src\IQS324\IQS324.h"

/*** Defines ***/
#define DEMO_IQS324_ADDR 0x54
#define DEMO_IQS324_POWER_PIN 4
#define DEMO_IQS324_RDY_PIN 7

/* UI Selection */
#define DEMO_HID_ON 1 // Turn on HID and give mouse scroll commands over USB.

#if DEMO_HID_ON != 0
#include <Mouse.h>
#endif

/*** Instances ***/
IQS324 iqs324; // Create an IQS324 instance

/*** Global Variables ***/
// Get the number of intervals from the IQS324_init settings file
const uint16_t MAX_INTERVALS = NUMBER_INTERVALS_0 + (NUMBER_INTERVALS_1 << 8);
uint8_t running_display_counter = 0;

void force_comms_and_reset(void);
char read_message(void);
void print_Heading(void);
void check_current_interval(void);
void check_button_state(void);
void check_power_mode(void);
void iqs324_ready_interrupt(void);
int16_t calculate_interval_diff(uint16_t new_interval, uint16_t old_interval);

void setup()
{
    /* Start Serial Communication */
    Serial.begin(115200);
    while (!Serial)
        ;
    Serial.println("Start Serial communication");

    /* Power On IQS324 */
    pinMode(DEMO_IQS324_POWER_PIN, OUTPUT);
    digitalWrite(DEMO_IQS324_POWER_PIN, LOW);
    delay(200);
    digitalWrite(DEMO_IQS324_POWER_PIN, HIGH);

    /* Initialize the IQS324 with input parameters device address and RDY pin */
    iqs324.begin(DEMO_IQS324_ADDR, DEMO_IQS324_RDY_PIN, iqs324_ready_interrupt);
    delay(1);

#if DEMO_HID_ON != 0
    Mouse.begin();
#endif
}

void loop()
{
    iqs324.run(); // Runs the IQS324

    force_comms_and_reset(); // function to initialize a force communication window.

    /* Process data read from IQS324 when new data is available (RDY Line Low) */
    if (iqs324.new_data_available)
    {
        print_Heading();
        check_current_interval();
        check_button_state();
        check_power_mode();
        iqs324.new_data_available = false;
    }
}

/* Function to check when the Current power mode of the IQS324 changed */
void check_power_mode(void)
{
    // read current power mode
    switch (iqs324.get_PowerMode())
    {
    case IQS324_HIGH_ACCURACY:
        Serial.println("HIGH ACCURACY");
        break;
    case IQS324_NORMAL_POWER:
        Serial.println("NORMAL POWER");
        break;
    case IQS324_LOW_POWER:
        Serial.println("LOW POWER");
        break;
    case IQS324_ULP:
        Serial.println("ULP");
        break;
    default:
        break;
    }
}

void print_Heading(void)
{
    /* Check if it is necessary to display the heading */
    if (running_display_counter == 11 || running_display_counter == 0)
    {
        /* Print the extra
        heading for the serial feedback */
        Serial.println("\nENC Interval:\tProx:\t\tPower Mode:");

        running_display_counter = 1;
    }

    running_display_counter++;
}

/* Function to check the Proximity and Touch states of the self-capacitance button */
void check_button_state(void)
{
    /* check if the movement state bit is set for channel 1 */
    if (iqs324.button_movementState())
    {
        Serial.print("Movement\t");
    }
    /* check if the button state bit is set for channel 1 */
    else if (iqs324.button_buttonState())
    {
        Serial.print("Button\t\t");
    }
    /* None state if neither the touch nor proximity bit was set to 1 */
    else
    {
        Serial.print("None\t\t");
    }
}

/* Check the current interval of the Quadrature */
void check_current_interval(void)
{

    uint16_t current_interval = iqs324.interval_value();
    int16_t interval_change = calculate_interval_diff(current_interval, iqs324.previous_interval);
    iqs324.previous_interval = current_interval;

    Serial.print(current_interval);
    Serial.print("\t\t");

#if DEMO_HID_ON != 0
    Mouse.move(0, 0, (int8_t)interval_change); // scroll mouse up by 1.
#endif
}

/* Force the IQS324 to open a RDY window and read the current state of the
device or request a software reset */
void force_comms_and_reset(void)
{
    char message = read_message();

    /* If an 'f' was received over serial, open a forced communication window and
    print the new data received */
    if (message == 'f')
    {
        iqs324.force_I2C_communication(); // prompt the IQS324
    }

    /* If an 'r' was received over serial, request a software(SW) reset */
    if (message == 'r')
    {
        Serial.println("Software Reset Requested!");
        iqs324.force_I2C_communication(); // Request a RDY window
        iqs324.do_soft_reset = true;
        running_display_counter = 0;
    }
}

/* Read message sent over serial communication */
char read_message(void)
{
    while (Serial.available())
    {
        if (Serial.available() > 0)
        {
            return Serial.read();
        }
    }

    return '\n';
}

void iqs324_ready_interrupt(void)
{
    iqs324.ready_interrupt();
}

/* Calculates the interval difference between two interval values.
This takes wrapping into account. */
int16_t calculate_interval_diff(uint16_t new_interval, uint16_t old_interval)
{
    int16_t diff = (int16_t)(new_interval - old_interval);
    int16_t half_intervals = (int16_t)(MAX_INTERVALS / 2);

    if (diff > half_intervals)
    {
        diff -= (int16_t)MAX_INTERVALS;
    }
    else if (diff < -half_intervals)
    {
        diff += (int16_t)MAX_INTERVALS;
    }

    return diff;
}
