/******************************************************************************
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        IQS324.cpp                                                  *
 * @brief       This file contains the constructors and methods which allow   *
 *              ease of use of an IQS324 Hall-rotation sensor. The IQS324 *
 *              is a capacitive touch and Hall-rotation Integrated Circuit    *
 *              (IC) which provides multiple-channel functionality. This      *
 *              class provides an easy means of initializing and interacting  *
 *              with the IQS324 device from an Arduino.                     *
 * @author      Azoteq PTY Ltd                                                *
 * @version     v1.0                                                          *
 * @date        2025                                                          *
 * ========================================================================== *
 * @attention   Makes use of the following standard Arduino libraries:        *
 * - Arduino.h -> Included in IQS324.h, comes standard with Arduino         *
 * - Wire.h    -> Included in IQS324.h, comes standard with Arduino         *
 ******************************************************************************/

/* Include Files */
#include "IQS324.h"

/******************************************************************************/
/*                                CONSTRUCTORS                                */
/******************************************************************************/
IQS324::IQS324()
{
}

/******************************************************************************/
/*                               PUBLIC METHODS                               */
/******************************************************************************/

/**
 * @name    begin
 * @brief   A method to initialize the IQS324 device with the device address
 *          and ready pin specified by the user.
 * @param   deviceAddress -> The address of the IQS324 device.
 * @param   readyPin      -> The Arduino pin which is connected to the ready
 *                           pin of the IQS324 device. Should be an
 *                           interrupt-capable pin.
 * @param   interrupt_func -> Pointer to function that must be executed on the
 *                            ready interrupt.
 *                            This is a user-defined function that should call
 *                            iqs7221g.ready_interrupt().
 * @retval  None.
 */
void IQS324::begin(uint8_t deviceAddressIn, uint8_t readyPinIn, void (*interrupt_func)())
{
    // Initialize I2C communication here, since this library can't function without it.
    Wire.begin();
    Wire.setClock(400000);

    _deviceAddress = deviceAddressIn;
    _readyPin = readyPinIn;
    _deviceRDY = false;
    _force_comms_rdy_received = false;
    do_soft_reset = false;
    attachInterrupt(digitalPinToInterrupt(_readyPin), interrupt_func, CHANGE);

    /* Initialize "running" and "init" state machine variables. */
    iqs324_state.state = IQS324_STATE_START;
    iqs324_state.init_state = IQS324_INIT_VERIFY_PRODUCT;
}

/**
 * @name    init
 * @brief   A method that runs through a normal start-up routine to set up the
 *          IQS324 with the desired settings from the IQS324_init.h file.
 * @details This function must be called repeatedly until it returns true.
 * @retval  Returns true if the full start-up routine has been completed,
 *          returns false if not.
 * @note    No false return will be given, the program will thus be stuck when
 *          one of the cases is not able to finish.
 *          See serial communication to find the ERROR case
 */
bool IQS324::init(void)
{
    uint16_t prod_num;
    uint8_t ver_maj, ver_min;

    switch (iqs324_state.init_state)
    {
    /* Verifies product number to determine if the correct device is connected for
    this example */
    case IQS324_INIT_VERIFY_PRODUCT:
        if (_deviceRDY)
        {
            Serial.println("\tIQS324_INIT_VERIFY_PRODUCT");
            prod_num = getProductNum(RESTART);
            ver_maj = getmajorVersion(RESTART);
            ver_min = getminorVersion(STOP);
            Serial.print("\t\tProduct number is: ");
            Serial.print(prod_num);
            Serial.print(" v");
            Serial.print(ver_maj);
            Serial.print(".");
            Serial.println(ver_min);
            if (prod_num == IQS324_PRODUCT_NUM)
            {
                Serial.println("\t\tIQS324 Confirmed!");
                iqs324_state.init_state = IQS324_INIT_READ_RESET;
            }
            else
            {
                Serial.println("\t\tDevice is not a IQS324!");
                iqs324_state.init_state = IQS324_INIT_NONE;
            }
        }
        break;

    /* Verify if a reset has occurred */
    case IQS324_INIT_READ_RESET:
        if (_deviceRDY)
        {
            Serial.println("\tIQS324_INIT_READ_RESET");
            updateInfoFlags(RESTART);
            if (checkReset())
            {
                Serial.println("\t\tReset event occurred.");
                iqs324_state.init_state = IQS324_INIT_UPDATE_SETTINGS;
            }
            else
            {
                Serial.println("\t\t No Reset Event Detected - Request SW Reset");
                iqs324_state.init_state = IQS324_INIT_CHIP_RESET;
            }
        }
        break;

    /* Perform SW Reset */
    case IQS324_INIT_CHIP_RESET:
        if (_deviceRDY)
        {
            Serial.println("\tIQS324_INIT_CHIP_RESET");

            // Perform SW Reset
            SW_Reset(STOP);
            Serial.println("\t\tSoftware Reset Bit Set.");
            delay(100);
            iqs324_state.init_state = IQS324_INIT_READ_RESET;
        }
        break;

    /* Write all settings to IQS324 from .h file */
    case IQS324_INIT_UPDATE_SETTINGS:
        if (_deviceRDY)
        {
            Serial.println("\tIQS324_INIT_UPDATE_SETTINGS");
            writeMM(RESTART);
            iqs324_state.init_state = IQS324_INIT_ACK_RESET;
        }
        break;

    /* Acknowledge that the device went through a reset */
    case IQS324_INIT_ACK_RESET:
        if (_deviceRDY)
        {
            Serial.println("\tIQS324_INIT_ACK_RESET");
            acknowledgeReset(STOP);
            iqs324_state.init_state = IQS324_INIT_ATI;
        }
        break;

    /* Run the ATI algorithm to recalibrate the device with newly added settings */
    case IQS324_INIT_ATI:
        if (_deviceRDY)
        {
            Serial.println("\tIQS324_INIT_ATI");
            ReATI(STOP);
            iqs324_state.init_state = IQS324_INIT_ZERO;
        }
        break;

    /* Zero the encoder angle of the iqs324 */
    case IQS324_INIT_ZERO:
        if (_deviceRDY)
        {
            zeroNow(STOP);
            Serial.println("\tIQS324_INIT_ZERO");
            iqs324_state.init_state = IQS324_INIT_READ_DATA;
        }
        break;

    /* Read the latest data from the iqs324 */
    case IQS324_INIT_READ_DATA:
        if (_deviceRDY)
        {
            Serial.println("\tIQS324_INIT_READ_DATA");
            queueValueUpdates();
            iqs324_state.init_state = IQS324_INIT_ACTIVATE_EVENT_MODE;
        }
        break;

    /* Turn on I2C Event mode */
    case IQS324_INIT_ACTIVATE_EVENT_MODE:
        if (_deviceRDY)
        {
            Serial.println("\tIQS324_INIT_ACTIVATE_EVENT_MODE");
            setEventMode(STOP);
            iqs324_state.init_state = IQS324_INIT_DONE;
        }
        break;

    /* All operations have been completed correctly */
    case IQS324_INIT_DONE:
        Serial.println("\tIQS324_INIT_DONE");
        new_data_available = true;
        return true;
        break;

    default:
        break;
    }
    return false;
}

/**
 * @name    run
 * @brief   Run the device ready check and read the necessary bytes when the
 *          IQS324 has pulled the RDY line low.
 *          The new_data_available flag will be set when a ready low is received
 *          from the IQS324.
 * @note    queueValueUpdates can be edited by the user if other data should be
 *          read every time a RDY window is received.
 */
void IQS324::run(void)
{
    switch (iqs324_state.state)
    {
    /* After a hardware reset, this is the starting position of the main running
    state machine */
    case IQS324_STATE_START:
        Serial.println("IQS324 Initialization:");
        iqs324_state.state = IQS324_STATE_INIT;
        break;

    /* Perform the initialization routine on the IQS324 */
    case IQS324_STATE_INIT:
        if (init())
        {
            Serial.println("IQS324 Initialization complete!\n");
            iqs324_state.state = IQS324_STATE_CHECK_RESET;
        }
        break;

    /* Continuous reset monitoring state, ensure no reset event has occurred
    for data to be valid */
    case IQS324_STATE_CHECK_RESET:
        if (checkReset())
        {
            Serial.println("Reset Occurred!\n");
            new_data_available = false;
            iqs324_state.state = IQS324_STATE_START;
            iqs324_state.init_state = IQS324_INIT_VERIFY_PRODUCT;
        }
        /* A reset did not occur, move to the run state and wait for a new ready
        window */
        else
        {
            new_data_available = true; /* No reset, thus data is valid */
            iqs324_state.state = IQS324_STATE_RUN;
        }
        break;

    /* When doing a force comms, add a short delay of 300 us after receiving
    the RDY to ensure that the comms window stays open. */
    case IQS324_STATE_WAIT_FORCE_COMMS:
        if (_force_comms_rdy_received && (micros() - force_comms_rdy_timestamp >= 300))
        {
            // After waiting 300 us, check the state of the RDY pin.
            // If it's high, the comms window closed early.
            // Re-attempt force comms.
            // If it's still low, go to the run state to handle comms normally.
            if (digitalRead(_readyPin))
            {
                force_I2C_communication();
                return;
            }
            else
            {
                iqs324_state.state = IQS324_STATE_RUN;
            }
        }
        break;

    /* If a RDY Window is open, read the latest values from the IQS324 */
    case IQS324_STATE_RUN:
        new_data_available = false;
        if (_deviceRDY)
        {
            if (do_soft_reset)
            {
                do_soft_reset = false;
                SW_Reset(STOP);
                return;
            }
            queueValueUpdates();
            _deviceRDY = false;
            iqs324_state.state = IQS324_STATE_CHECK_RESET;
        }
        break;

    default:
        break;
    }
}

/**
 * @name    iqs324_ready_interrupt
 * @brief   A method used as an interrupt function. Only activated when a falling-
 *          edge interrupt is seen on the correct Arduino interrupt pin.
 * @note    Keep this function as simple as possible to prevent stuck states and
 *          slow operations.
 */
void IQS324::ready_interrupt(void)
{
    if (digitalRead(_readyPin))
    {
        _deviceRDY = false;
    }
    else
    {
        _deviceRDY = true;
        // If a force comms was requested, store the timestamp of when the RDY changed.
        // Will then wait 300 us and re-check the state of the RDY pin to make
        // sure the force comms window has stayed open.
        if (iqs324_state.state == IQS324_STATE_WAIT_FORCE_COMMS)
        {
            force_comms_rdy_timestamp = micros();
            _force_comms_rdy_received = true;
        }
    }
}

/**
 * @name   clearRDY
 * @brief  A method used to clear the ready interrupt bit.
 */
void IQS324::clearRDY(void)
{
    _deviceRDY = false;
}

/**
 * @name   getRDYStatus
 * @brief  A method used to retrieve the device RDY status.
 * @param  None.
 * @retval Returns the boolean IQS323 RDY state.
 *         - True when RDY line is LOW
 *         - False when RDY line is HIGH
 */
bool IQS324::getRDYStatus(void)
{
    return _deviceRDY;
}

/**
 * @name    queueValueUpdates
 * @brief   All I2C read operations in the queueValueUpdates method will be
 *          performed each time the IQS324 opens a RDY window.
 * @note    Any Address in the memory map can be read from here.
 */
void IQS324::queueValueUpdates(void)
{

    uint8_t transferBytes[10]; // The array which will hold the bytes to be transferred.

    /* Read the information necessary for this specific implementation from the
    IQS324 */
    readRandomBytes(IQS324_MM_SYSTEM_FLAGS, 8, transferBytes, RESTART);

    /* Assign the system status flags to the local device_status variable */
    IQSMemoryMap.power_mode = transferBytes[0];
    IQSMemoryMap.device_status = transferBytes[1];

    /* Assign the HALL UI Information to the local hall_flags variable */
    IQSMemoryMap.hall_flags = transferBytes[2];

    /* Assign the events flags to the local events variables */
    IQSMemoryMap.event_flags = transferBytes[3];
    IQSMemoryMap.prox_states = transferBytes[4];

    IQSMemoryMap.button_flags = transferBytes[6];
    IQSMemoryMap.movement_flags = transferBytes[7];

    /* Read the hall wheel interval. */
    readRandomBytes(IQS324_MM_HALL_INTERVAL, 2, transferBytes, STOP);

    /* Assign the hall UI interval into the local hall_ui_interval variable. */
    IQSMemoryMap.hall_ui_interval[0] = transferBytes[0];
    IQSMemoryMap.hall_ui_interval[1] = transferBytes[1];
}

/**
 * @name	checkReset
 * @brief   A method that checks if the device has reset and returns the reset
 *          status.
 * @retval  Returns true if a reset has occurred, false if no reset has
 *          occurred.
 * @note    If a reset has occurred the device settings should be reloaded using
 *          the begin function. After new device settings have been reloaded the
 *          acknowledge reset function can be used to clear the reset flag.
 */
bool IQS324::checkReset(void)
{
    /* Perform a bitwise AND operation inside getBIT with the DEVICE_RESET_BIT
   to return the reset status */
    return getBit(IQSMemoryMap.device_status, IQS324_SHOW_RESET_BIT);
}

/**
 * @name	checkProductNum
 * @brief   Read and return the device product number.
 * @param   stopOrRestart -> Specifies whether the communications window must
 *                           be kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @retval  Returns the product number as a 16-bit unsigned integer value.
 * @note    If the product is not correctly identified an appropriate messages
 *          should be displayed.
 */
uint16_t IQS324::getProductNum(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // A temporary array to hold the read bytes
    uint16_t prodNumReturn;   // The 16-bit return value.

    // Read the Device info from the IQS324.
    readRandomBytes(IQS324_MM_PROD_NUM, 2, transferBytes, stopOrRestart);

    // Construct the 16-bit return value.
    prodNumReturn = (uint16_t)(transferBytes[0]);
    prodNumReturn |= (uint16_t)(transferBytes[1] << 8);

    // Return the product number
    return prodNumReturn;
}

/**
 * @name	getmajorVersion
 * @brief   A method that reads the device firmware major version number.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @retval Returns the major version number as an 8-bit unsigned integer value.
 */
uint8_t IQS324::getmajorVersion(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // A temporary array to hold the read bytes

    // Read the Device info from the IQS324.
    readRandomBytes(IQS324_MM_MAJOR_VERSION_NUM, 2, transferBytes, stopOrRestart);

    // Return the major firmware version number value.
    return transferBytes[0];
}

/**
 * @name	getminorVersion
 * @brief   A method that reads the device firmware minor version number.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @retval Returns the minor version number as an 8-bit unsigned integer value.
 */
uint8_t IQS324::getminorVersion(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // A temporary array to hold the read bytes

    // Read the Device info from the IQS324.
    readRandomBytes(IQS324_MM_MINOR_VERSION_NUM, 2, transferBytes, stopOrRestart);

    // Return the minor firmware version number value.
    return transferBytes[0];
}

/**
 * @name	acknowledgeReset
 * @brief   A method that clears the Show Reset flag by setting the ACK_RESET
 *          bit.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    If a reset has occurred the device settings should be reloaded using
 *          the begin function. After new device settings have been reloaded,
 *          this method should be used to clear the reset bit.
 * @note    To acknowledge reset, IQS324_ACK_RESET_BIT of SYSTEM SETTINGS
 *          is set.
 */
void IQS324::acknowledgeReset(bool stopOrRestart)
{
    uint8_t transferBytes[1]; // A temporary array to hold the bytes transferred.

    // Read the System Flags from the IQS324
    // These must be read first in order not to change any settings.
    readRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 1, transferBytes, RESTART);
    // Write the Ack Reset bit to 1 to clear the Show Reset Flag.
    transferBytes[0] = setBit(transferBytes[0], IQS324_ACK_RESET_BIT);
    // Write the new byte to the System Flags address.
    writeRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 1, transferBytes, stopOrRestart);
}

/**
 * @name    ReATI
 * @brief   A method that sets the REDO_ATI_BIT in order to force the IQS324
 *          device to run the Automatic Tuning Implementation (ATI) routine.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To perform ATI, bits 2 and 3 of SYSTEM SETTINGS is set.
 */
void IQS324::ReATI(bool stopOrRestart)
{
    uint8_t transferByte[1]; // Array to store the bytes transferred.

    readRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 1, transferByte, RESTART);
    // Mask the settings with the REDO_ATI_BIT.
    transferByte[0] = setBit(transferByte[0], IQS324_FORCE_ATI_HALL_BIT);
    transferByte[0] = setBit(transferByte[0], IQS324_FORCE_ATI_PF_BIT);
    // Write the new byte to the required device.
    writeRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 1, transferByte, stopOrRestart);
}

/**
 * @name    SW_Reset
 * @brief   A method that sets the SW RESET bit to force the IQS324
 *          device to do a SW reset.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To perform SW Reset, bit 1 in SYSTEM SETTINGS is set.
 */
void IQS324::SW_Reset(bool stopOrRestart)
{
    uint8_t transferByte[1]; // Array to store the bytes transferred.

    readRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 1, transferByte, RESTART);
    // Mask the settings with the SW_RESET_BIT.
    transferByte[0] = setBit(transferByte[0], IQS324_SW_RESET_BIT);
    // Write the new byte to the required device.
    writeRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 1, transferByte, stopOrRestart);
}

/**
 * @name    setEventMode
 * @brief   A method to set the IQS324 device into event mode.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To enter event mode, IQS324_EVENT_MODE_BIT of SYSTEM SETTINGS
 *          is set.
 */
void IQS324::setEventMode(bool stopOrRestart)
{
    uint8_t transferBytes[1]; // The array which will hold the bytes which are transferred.

    // First read the bytes at the memory address so that they can be preserved.
    readRandomBytes(IQS324_MM_SYSTEM_SETTINGS_1, 1, transferBytes, RESTART);

    // Set the EVENT_MODE_BIT in UI_SETTINGS
    transferBytes[0] = setBit(transferBytes[0], IQS324_EVENT_MODE_BIT);
    // Write the bytes back to the device
    writeRandomBytes(IQS324_MM_SYSTEM_SETTINGS_1, 1, transferBytes, stopOrRestart);
}

/**
 * @name    setStreamMode
 * @brief   A method to set the IQS324 device into streaming mode.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To enter event mode, IQS324_EVENT_MODE_BIT of SYSTEM SETTINGS
 *          is cleared.
 */
void IQS324::setStreamMode(bool stopOrRestart)
{
    uint8_t transferBytes[1]; // The array which will hold the bytes transferred.

    // First read the bytes at the memory address so that they can be preserved.
    readRandomBytes(IQS324_MM_SYSTEM_SETTINGS_1, 1, transferBytes, RESTART);
    // Clear the EVENT_MODE bit in SYSTEM SETTINGS
    transferBytes[0] = clearBit(transferBytes[0], IQS324_EVENT_MODE_BIT);
    // Write the bytes back to the device
    writeRandomBytes(IQS324_MM_SYSTEM_SETTINGS_1, 1, transferBytes, stopOrRestart);
}

/**
 * @name    zeroNow
 * @brief   A method to zero the encoder angle of the IQS324.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To run the zero-now command, set IQS324_ZERO_NOW_BIT of
 *          SYSTEM SETTINGS.
 */
void IQS324::zeroNow(bool stopOrRestart)
{
    uint8_t transferBytes[1]; // The array which will hold the bytes transferred.

    // First read the bytes at the memory address so that they can be preserved.
    readRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 1, transferBytes, RESTART);
    // Set the ZERO_NOW_BIT in SYSTEM SETTINGS
    transferBytes[0] = setBit(transferBytes[0], IQS324_ZERO_NOW_BIT);
    // Write the bytes back to the device
    writeRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 1, transferBytes, stopOrRestart);
}

/**
 * @name    updateInfoFlags
 * @brief   A method which reads the IQS324 info flags and assigns them to the
 *          local IQSMemoryMap struct.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    The local `power_mode` and `device_status` values are altered with
 *          the new value of the info flags registers retrieved from the IQS324.
 */
void IQS324::updateInfoFlags(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // The array which will hold the bytes to be transferred.

    // Read the info flags.
    readRandomBytes(IQS324_MM_SYSTEM_FLAGS, 2, transferBytes, stopOrRestart);
    // Assign the info flags to the local SYSTEM_STATUS register.
    IQSMemoryMap.power_mode = transferBytes[0];
    IQSMemoryMap.device_status = transferBytes[1];
}

/**
 * @name    get_PowerMode
 * @brief   A method that reads the local `power_mode` register and returns
 *          the current power mode.
 * @retval  Returns the iqs324_power_modes state the device is in.
 * @note    See Datasheet on power mode options and timeouts.
 *          High Accuracy, Normal Power, Low Power, Ultra Low Power (ULP).
 */
IQS324_power_modes IQS324::get_PowerMode(void)
{
    uint8_t buffer = getBit(IQSMemoryMap.power_mode, IQS324_POWER_MODE_BIT_0);
    buffer += getBit(IQSMemoryMap.power_mode, IQS324_POWER_MODE_BIT_1) << 1;

    return (IQS324_power_modes)buffer;
}

/**
 * @name    button_movementState
 * @brief   A method that checks the local `movement_flags` if the button is
 *          in a movement active state.
 * @retval  Returns true if a movement is detected and false if there is no movement.
 * @note    The IQS324 only has 1 touch output channel.
 */
bool IQS324::button_movementState(void)
{
    return getBit(IQSMemoryMap.movement_flags, IQS324_MOVEMENT_ACTIVE_STATE_BIT);
}

/**
 * @name    button_buttonState
 * @brief   A method that checks the local `button_flags` register if the
 *          button is in a touch state.
 * @retval  Returns true if touch is active and false if there is no
 *          touch.
 * @note    The IQS324 only has 1 touch output channel.
 */
bool IQS324::button_buttonState(void)
{
    return getBit(IQSMemoryMap.button_flags, IQS324_BUTTON_ACTIVE_STATE_BIT);
}

/**
 * @name    interval_changed
 * @brief   A method that checks the Hall UI flags if there was a change in
 *          interval.
 * @retval  Returns true if an interval change occurred.
 */
bool IQS324::interval_changed(void)
{
    return getBit(IQSMemoryMap.hall_flags, IQS324_HALL_INTERVAL_CHANGED_BIT);
}

/**
 * @name    interval_direction
 * @brief   A method that reads the local HALL_UI register for the Hall UI flags
 *          and determines in what direction the magnet rotated.
 * @retval  Returns true if a positive interval change occurred.
 */
bool IQS324::interval_direction(void)
{
    return getBit(IQSMemoryMap.hall_flags, IQS324_HALL_DIRECTION_BIT);
}

/**
 * @name    interval_value
 * @brief   A method that reads and returns the current interval value.
 * @retval  Returns the 16-bit unsigned integer value of the current
 *          interval.
 */
uint16_t IQS324::interval_value(void)
{
    return (IQSMemoryMap.hall_ui_interval[0] + (IQSMemoryMap.hall_ui_interval[1] << 8));
}

/******************************************************************************/
/*                           ADVANCED PUBLIC METHODS                          */
/******************************************************************************/

/**
 * @name    writeMM
 * @brief   Function to write the whole memory map to the device (writable)
 *          registers
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    Requires the `IQS324_init.h` file with the defined settings.
 */
void IQS324::writeMM(bool stopOrRestart)
{
    // Temporary array which holds the bytes to be transferred.
    uint8_t transferBytes[32];

    /* Change the System Control Settings */
    /* Memory Map Position 0x1000 - 0x1011 */
    transferBytes[0] = AUTO_CLEAR_SETTINGS;
    transferBytes[1] = SETTINGS_0;
    transferBytes[2] = EVENTMASK;
    transferBytes[3] = ULP_SETTINGS;
    transferBytes[4] = AUTO_ULP_THRESHOLD_0;
    transferBytes[5] = AUTO_ULP_THRESHOLD_1;
    transferBytes[6] = HA_REPORT_0;
    transferBytes[7] = HA_REPORT_1;
    transferBytes[8] = NP_REPORT_0;
    transferBytes[9] = NP_REPORT_1;
    transferBytes[10] = LP_REPORT_0;
    transferBytes[11] = LP_REPORT_1;
    transferBytes[12] = ULP_REPORT_0;
    transferBytes[13] = ULP_REPORT_1;
    transferBytes[14] = ULP_TIMEOUT_0;
    transferBytes[15] = ULP_TIMEOUT_1;
    transferBytes[16] = I2C_TIMEOUT_0;
    transferBytes[17] = I2C_TIMEOUT_1;

    writeRandomBytes(IQS324_MM_SYSTEM_SETTINGS, 18, transferBytes, RESTART);
    Serial.println("\t\t1. Write System Settings");

    /* Change the Channel 0 Settings */
    /* Memory Map Position 0x1014 - 0x102D */
    transferBytes[0] = CONVERSION_SETUP;
    transferBytes[1] = CONVERSION_SETTINGS;
    transferBytes[2] = BUTTON_TOUCH_TIMEOUT_0;
    transferBytes[3] = BUTTON_TOUCH_TIMEOUT_1;
    transferBytes[4] = BUTTON_DORMANCY_TIMEOUT_0;
    transferBytes[5] = BUTTON_DORMANCY_TIMEOUT_1;
    transferBytes[6] = BUTTON_ATI_TIMEOUT_0;
    transferBytes[7] = BUTTON_ATI_TIMEOUT_1;
    transferBytes[8] = BUTTON_FILTER_COUNT_BETA;
    transferBytes[9] = BUTTON_FILTER_LTA_BETA;
    transferBytes[10] = BUTTON_FILTER_LTA_BETA_FAST;
    transferBytes[11] = BUTTON_FILTER_LTA_LP_BETA;
    transferBytes[12] = BUTTON_FILTER_LTA_BETA_FAST_BAND_0;
    transferBytes[13] = BUTTON_FILTER_LTA_BETA_FAST_BAND_1;
    transferBytes[14] = BUTTON_THRESHOLD_0;
    transferBytes[15] = BUTTON_THRESHOLD_1;
    transferBytes[16] = BUTTON_DEBOUNCE;
    transferBytes[17] = BUTTON_PROX_HYST;
    transferBytes[18] = MOVEMENT_THRESHOLD_0;
    transferBytes[19] = MOVEMENT_THRESHOLD_1;
    transferBytes[20] = MOVEMENT_DEBOUNCE;
    transferBytes[21] = MOVEMENT_TOUCH_HYST;
    transferBytes[22] = BUTTON_ATI_BAND_0;
    transferBytes[23] = BUTTON_ATI_BAND_1;
    transferBytes[24] = BUTTON_LTA_HALT_THRESHOLD_0;
    transferBytes[25] = BUTTON_LTA_HALT_THRESHOLD_1;

    writeRandomBytes(IQS324_MM_CHANNEL0_SETTINGS, 26, transferBytes, RESTART);
    Serial.println("\t\t1. Write Channel 0 Settings");

    /* Change the Channel 0 ATI Settings */
    /* Memory Map Position 0x102E - 0x1035 */
    transferBytes[0] = SENSOR0_ATI_BASE_0;
    transferBytes[1] = SENSOR0_ATI_BASE_1;
    transferBytes[2] = SENSOR0_ATI_TARGET_0;
    transferBytes[3] = SENSOR0_ATI_TARGET_1;
    transferBytes[4] = SENSOR0_COARSE;
    transferBytes[5] = SENSOR0_FINE;
    transferBytes[6] = SENSOR0_COMPENSATION;
    transferBytes[7] = SENSOR0_COMPENSATION_DIVIDER;

    writeRandomBytes(IQS324_MM_CHANNEL0_ATI_SETTINGS, 8, transferBytes, RESTART);
    Serial.println("\t\t1. Write Channel ATI Settings");

    /* Change the Hall Plate Settings */
    /* Memory Map Position 0x1036 - 0x1047 */
    transferBytes[0] = NW_HALLPLATE_OFFSET_0;
    transferBytes[1] = NW_HALLPLATE_OFFSET_1;
    transferBytes[2] = NE_HALLPLATE_OFFSET_0;
    transferBytes[3] = NE_HALLPLATE_OFFSET_1;
    transferBytes[4] = SE_HALLPLATE_OFFSET_0;
    transferBytes[5] = SE_HALLPLATE_OFFSET_1;
    transferBytes[6] = SW_HALLPLATE_OFFSET_0;
    transferBytes[7] = SW_HALLPLATE_OFFSET_1;
    transferBytes[8] = HALL_COARSE;
    transferBytes[9] = HALL_FINE;
    transferBytes[10] = HALL_TARGET_0;
    transferBytes[11] = HALL_TARGET_1;
    transferBytes[12] = HALL_ATI_BAND_0;
    transferBytes[13] = HALL_ATI_BAND_1;
    transferBytes[14] = HP_BIAS;
    transferBytes[15] = HP_SETTING;
    transferBytes[16] = RE_ATI_BLOCKING;
    transferBytes[17] = HALL_INIT_SETTING;
    writeRandomBytes(IQS324_MM_HALL_PLATE_SETTINGS, 18, transferBytes, RESTART);
    Serial.println("\t\t1. Write Hall PLate Settings");

    /* Change the Hall Phase Angle */
    /* Memory Map Position 0x1048 - 0x104B */
    transferBytes[0] = PHASE_SINE_0;
    transferBytes[1] = PHASE_SINE_1;
    transferBytes[2] = PHASE_COSE_0;
    transferBytes[3] = PHASE_COSE_1;
    writeRandomBytes(IQS324_HALL_PHASE_ANGLE_SETTINGS, 4, transferBytes, RESTART);
    Serial.println("\t\t1. Write Hall Phase Angle Settings");

    /* Change the Hall UI Settings */
    /* Memory Map Position 0x1062 - 0x106D */
    transferBytes[0] = ROTATION_UI_SETTINGS_0;
    transferBytes[1] = OPEN_QUAD;
    transferBytes[2] = WHEEL_TO_MAGNET_ANGLE_0;
    transferBytes[3] = WHEEL_TO_MAGNET_ANGLE_1;
    transferBytes[4] = HIGH_ACCURACY_TIMEOUT_0;
    transferBytes[5] = HIGH_ACCURACY_TIMEOUT_1;
    transferBytes[6] = STATIONARY_TIMEOUT_0;
    transferBytes[7] = STATIONARY_TIMEOUT_1;
    transferBytes[8] = AUTO_ZERO_BETA;
    transferBytes[9] = FILTER_BETA;
    transferBytes[10] = LP_FILTER_BETA;
    transferBytes[11] = CORDIC_RESOLUTION;
    writeRandomBytes(IQS324_MM_HALL_UI_SETTINGS, 12, transferBytes, RESTART);
    Serial.println("\t\t1. Write Hall UI Settings");

    /* Change the Hall Interval Settings */
    /* Memory Map Position 0x106E - 0x1073 */
    transferBytes[0] = NUMBER_INTERVALS_0;
    transferBytes[1] = NUMBER_INTERVALS_1;
    transferBytes[2] = INTERVAL_HYSTERESIS_0;
    transferBytes[3] = INTERVAL_HYSTERESIS_1;
    transferBytes[4] = DEFAULT_HYSTERESIS_0;
    transferBytes[5] = DEFAULT_HYSTERESIS_1;
    writeRandomBytes(IQS324_MM_HALL_INTERVAL_SETTINGS, 6, transferBytes, RESTART);
    Serial.println("\t\t1. Write Hall Interval Settings");

    /* Change the Normalisation Settings */
    /* Memory Map Position 0x1074 - 0x1075 */
    transferBytes[0] = NORMILIZATION_BETA;
    transferBytes[1] = NORMILIZATION_MODE;
    writeRandomBytes(IQS324_MM_NORMALISATION_SETTINGS, 2, transferBytes, stopOrRestart);
    Serial.println("\t\t1. Normalisation Settings");
}

/******************************************************************************/
/*                               PRIVATE METHODS                              */
/******************************************************************************/

/**
 * @name    readRandomBytes
 * @brief   A method that reads a specified number of bytes from a specified
 *          16-bit address and saves it into a user-supplied array.
 * @param   memoryAddress -> The memory address at which to start reading bytes
 *                           from.
 * @param   numBytes      -> The number of bytes that must be read.
 * @param   bytesArray    -> The array which will store the bytes to be read,
 *                           this array will be overwritten.
 * @param   stopOrRestart -> A boolean that specifies whether the communication
 *                           window should remain open or be closed after transfer.
 *                           False keeps it open, true closes it. Use the STOP
 *                           and RESTART definitions.
 * @retval  No value is returned, however, the user-supplied array is overwritten.
 * @note    Uses standard Arduino "Wire" library which is for I2C communication.
 *          Take note that C++ cannot return an array, therefore, the array which
 *          is passed as an argument is overwritten with the required values.
 *          Pass an array to the method by using only its name, e.g. "bytesArray",
 *          without the brackets, this passes a pointer to the array.
 */

void IQS324::readRandomBytes(uint16_t memoryAddress,
                             uint8_t numBytes,
                             uint8_t bytesArray[],
                             bool stopOrRestart)
{
    int i = 0;
    uint8_t error_s = 0; // I2C Error variable to track any issues during communication

    /* Select the device with the address "_deviceAddress" and start
    communication. */
    Wire.beginTransmission(_deviceAddress);
    // Specify the memory address where the IQS324 must start saving the data,
    // as designated by the "memoryAddress" variable.
    uint16_t addr_h = memoryAddress >> 8;
    uint16_t addr_l = memoryAddress;

    Wire.write(addr_l);
    Wire.write(addr_h);

    error_s = Wire.endTransmission(RESTART); // Restart transmission for reading that follows.
    /* The required device has now been selected and it has been told which register to send information from. */

    /* Request "numBytes" bytes from the device which has address "deviceAddress"*/
    uint8_t counter = 0;
    do
    {
        Wire.requestFrom((int)_deviceAddress, (int)numBytes, (int)stopOrRestart);

        /* break out of request loop if max retry is reached */
        if (counter++ >= IQS324_I2C_RETRY)
        {
            return;
        }
    } while (Wire.available() == 0); // Wait for response, this sometimes takes a few attempts

    /* Load the received bytes into the array until there are no more */
    while (Wire.available())
    {
        /* Load the received bytes into the user supplied array */
        bytesArray[i] = Wire.read();
        i++;
    }

    /* Always manually close the RDY window after a STOP is sent to prevent
    writing while the RDY window closes */
    if (stopOrRestart == STOP)
    {
        _deviceRDY = false;
    }
}

/**
 * @name   writeRandomBytes
 * @brief  A method that writes a specified number of bytes to a specified
 *         16-bit address, the bytes to write are supplied using an array pointer.
 * @param  memoryAddress -> The memory address at which to start writing the
 *         bytes to.
 * @param  numBytes      -> The number of bytes that must be written.
 * @param  bytesArray    -> The array which stores the bytes which will be
 *                          written to the memory location.
 * @param  stopOrRestart -> A boolean that specifies whether the communication
 *                          window should remain open or be closed of transfer.
 *                          False keeps it open, true closes it. Use the STOP
 *                          and RESTART definitions.
 * @retval No value is returned, only the IQS device registers are altered.
 * @note   Uses standard Arduino "Wire" library which is for I2C communication.
 *         Take note that a full array cannot be passed to a function in C++.
 *         Pass an array to the function by using only its name, e.g. "bytesArray",
 *         without the square brackets, this passes a pointer to the
 *         array. The values to be written must be loaded into the array prior
 *         to passing it to the function.
 */
void IQS324::writeRandomBytes(uint16_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart)
{
    uint8_t error_s = 0; // I2C Error variable to track any issues during communication

    /* Select the device with the address of "deviceAddress" and start communication. */
    Wire.beginTransmission(_deviceAddress);

    /* Specify the memory address */
    uint16_t addr_h = memoryAddress >> 8;
    uint16_t addr_l = memoryAddress;

    Wire.write(addr_l);
    Wire.write(addr_h);

    /* Write the bytes as specified in the array. */
    for (int i = 0; i < numBytes; i++)
    {
        Wire.write(bytesArray[i]);
    }
    /* End the transmission, user decides to STOP or RESTART. */
    error_s = Wire.endTransmission(stopOrRestart);

    /* Always manually close the RDY window after a STOP is sent to prevent
    writing while the RDY window closes */
    if (stopOrRestart == STOP)
    {
        _deviceRDY = false;
    }
}

/**
 * @name   getBit
 * @brief  A method that returns the chosen bit value of the provided byte.
 * @param  data       -> byte of which a given bit value needs to be calculated.
 * @param  bit_number -> a number between 0 and 7 representing the bit in question.
 * @retval The boolean value of the specific bit requested.
 */
bool IQS324::getBit(uint8_t data, uint8_t bit_number)
{
    return (data & (1 << bit_number)) >> bit_number;
}

/**
 * @name   setBit
 * @brief  A method that sets the chosen bit value of the provided byte.
 * @param  data       -> byte of which a given bit value needs to be calculated.
 * @param  bit_number -> a number between 0 and 7 representing the bit in question.
 * @retval Returns an 8-bit unsigned integer value of the given data byte with
 *         the requested bit set.
 */
uint8_t IQS324::setBit(uint8_t data, uint8_t bit_number)
{
    return (data |= 1UL << bit_number);
}

/**
 * @name   clearBit
 * @brief  A method that clears the chosen bit value of the provided byte.
 * @param  data       -> byte of which a given bit value needs to be calculated.
 * @param  bit_number -> a number between 0 and 7 representing the bit in question.
 * @retval Returns an 8-bit unsigned integer value of the given data byte with
 *         the requested bit cleared.
 */
uint8_t IQS324::clearBit(uint8_t data, uint8_t bit_number)
{
    return (data &= ~(1UL << bit_number));
}

/**
 * @name   force_I2C_communication
 * @brief  A method that writes 0xFF to open a communication window on the
 *         IQS324.
 * @note   Uses standard Arduino "Wire" library which is for I2C communication.
 */
void IQS324::force_I2C_communication(void)
{
    /*Ensure RDY is HIGH at the moment*/
    if (digitalRead(_readyPin))
    {
        _force_comms_rdy_received = false;
        iqs324_state.state = IQS324_STATE_WAIT_FORCE_COMMS;

        /* Select the device with the address: "DEMO_IQS324_ADDR" and start
        communication. */
        Wire.beginTransmission(_deviceAddress);

        /* Write to memory address 0xFF that will prompt the IQS324 to open a
        communication window.*/
        Wire.write(0xFF);

        /* End the transmission, user decides to STOP or RESTART. */
        Wire.endTransmission(STOP);
    }
}
